#include "hierachy.h"
#include "main.h"

#include "menuroot.h"
#include "toolbarroot.h"
#include "keysroot.h"
#include "prefsroot.h"
#include "winoptroot.h"
#include "themesroot.h"

#define	PIXMAP_SPACING			4

IceWM_CM_NAMESPACE

/**
 * Hierachy::init:
 *
 * Creates the hierachical view.
 */
GtkWidget* Hierachy::init (void) {
    widget = GTK_CTREE (gtk_ctree_new (1, 0));

    gtk_clist_set_selection_mode (GTK_CLIST (widget), GTK_SELECTION_BROWSE);
    gtk_container_set_border_width (GTK_CONTAINER (widget), 3);
    
    gtk_signal_connect (GTK_OBJECT (widget), "tree-select-row", 
    			GTK_SIGNAL_FUNC (&tree_select_row), this);
    gtk_signal_connect (GTK_OBJECT (widget), "tree-unselect-row", 
    			GTK_SIGNAL_FUNC (&tree_unselect_row), this);
    gtk_signal_connect_after (GTK_OBJECT (widget), "tree-move", 
    			      GTK_SIGNAL_FUNC (&tree_move), this);
/*			      
    gtk_signal_connect_after (GTK_OBJECT (widget), "tree-expand", 
    			      GTK_SIGNAL_FUNC (&update_movement_rules), this);
    gtk_signal_connect_after (GTK_OBJECT (widget), "tree-collapse", 
    			      GTK_SIGNAL_FUNC (&update_movement_rules), this);
*/

    gtk_ctree_select (widget,
    insert_tree_node (menu = new MenuRoot ()));
    insert_tree_node (toolbar = new ToolbarRoot ());
    insert_tree_node (keys = new KeysRoot ());
    insert_tree_node (prefs = new PrefsRoot ());
    insert_tree_node (winopts = new WinOptRoot ());
//    insert_tree_node (themes = new ThemesRoot ());

    load (RootNode::LOC_PRIVATE);

    return GTK_WIDGET (widget);
}

/**
 * Hierachy::delete_tree_node:
 *
 * Destructs the associated TreeNode if a GtkCTreeNode is destroyed.
 */
void Hierachy::delete_tree_node (TreeNode*	node) {
    g_return_if_fail (node != NULL);
    delete node;
}

/**
 * Hierachy::insert_tree_node:
 *
 * Adds a configuration node to the hierachical view.
 */
GtkCTreeNode* Hierachy::insert_tree_node (TreeNode*	node,
                                          GtkCTreeNode*	parent,
                                          GtkCTreeNode*	sibling) {
    GtkCTreeNode*		new_node;
    gchar*			row [1];
    const TreeNode::Icon*	icon;
    
    g_assert (node);
    g_assert (widget);
    
    *row = (gchar*) node->get_label ();
    icon = node->get_icon ();
    
    g_return_val_if_fail (*row != NULL, 0);
    
    new_node = gtk_ctree_insert_node (widget, parent, sibling, row,
    				      PIXMAP_SPACING,
                                      icon ? icon->small_pixmap : 0,
                                      icon ? icon->small_bitmap : 0,
				      0, 0, node->is_leaf (),
				      node->insert_expanded ());
    gtk_ctree_node_set_row_data_full (widget, new_node, node,
                                      (GtkDestroyNotify) delete_tree_node);
    node->create_sub_tree (this, new_node);
    
    return new_node;
}

/**
 * Hierachy::insert_interactively:
 *
 * Adds a behind the currently selected node.
 */
void Hierachy::insert_interactively (TreeNode*	node) {
    GtkCTreeNode* curr_node = current ();
    GtkCTreeRow* curr_row = GTK_CTREE_ROW (curr_node);

    gtk_clist_freeze (GTK_CLIST (widget));
    
    if (curr_row->is_leaf || (tree_node (curr_node)->is_moveable () && 
        !curr_row->expanded && curr_row->children))
        gtk_ctree_select (widget,
            insert_tree_node (node, curr_row->parent, curr_row->sibling));
    else {
        gtk_ctree_select (widget,
            insert_tree_node (node, curr_node, curr_row->children));
        gtk_ctree_expand (widget, curr_node);
    }

    gtk_clist_thaw (GTK_CLIST (widget));
}

/**
 * Hierachy::remove_tree_node:
 *
 * Removes a tree node to the hierachical view.
 */
void Hierachy::remove_tree_node (TreeNode* node) {
    GtkCTreeNode* ctree_node;

    ctree_node = 
          node ? ctree_node = gtk_ctree_find_by_row_data (widget, 0, node)
               : current ();

    g_return_if_fail (ctree_node != NULL);
    gtk_ctree_remove_node (widget, ctree_node);
}

/**
 * Hierachy::remove_tree_node_children:
 *
 * Removes a the children of a tree node in the hierachical view.
 */
void Hierachy::remove_tree_node_children (TreeNode* node = 0) {
    GtkCTreeNode* ctree_node;

    ctree_node = 
          node ? ctree_node = gtk_ctree_find_by_row_data (widget, 0, node)
               : current ();

    g_return_if_fail (ctree_node != NULL);
    gtk_clist_freeze (GTK_CLIST (widget));
    while (GTK_CTREE_ROW (ctree_node)->children)
        gtk_ctree_remove_node (widget, GTK_CTREE_ROW (ctree_node)->children);
    gtk_clist_thaw (GTK_CLIST (widget));
}
    
/**
 * Hierachy::update_tree_node:
 *
 * Updates configuration node in the hierachical view.
 */
void Hierachy::update_tree_node (GtkCTreeNode*	row) {
    update_tree_node (tree_node (row ? row : current ()), 
                                 row ? row : current ());
};
				 
/**
 * Hierachy::update_tree_node:
 *
 * Updates configuration node in the hierachical view.
 */
void Hierachy::update_tree_node (TreeNode*	node,
                                 GtkCTreeNode*	row) {
    g_return_if_fail (node != NULL);

    if (!row)
        row = gtk_ctree_find_by_row_data (widget, 0, node);
	
    g_return_if_fail (row != NULL);
    
    const TreeNode::Icon* icon = node->get_icon ();
    gtk_ctree_node_set_pixtext (widget, row, 0, node->get_label (),
    				PIXMAP_SPACING,
                                icon ? icon->small_pixmap : 0,
                                icon ? icon->small_bitmap : 0);
}

/**
 * Hierachy::prev_node_is_moveable:
 *
 * Checks if the visually previous node is moveable.
 */
bool Hierachy::prev_node_is_moveable (GtkCTreeNode* ctree_node) {
    if (!ctree_node) ctree_node = current ();
    g_return_val_if_fail (ctree_node, false);
	
    TreeNode* node = tree_node (ctree_node);
    g_return_val_if_fail (node, false);
    
    GtkCTreeNode* prev_node;
    return  (node->is_moveable ()) &&
            (prev_node = GTK_CTREE_NODE_PREV (ctree_node)) &&
	    (tree_node (prev_node)->is_moveable ());
}
    
/**
 * Hierachy::next_node_is_moveable:
 *
 * Checks if the visually sibling node is moveable.
 */
bool Hierachy::next_node_is_moveable (GtkCTreeNode *ctree_node) {
    if (!ctree_node) ctree_node = current ();
    g_return_val_if_fail (ctree_node, false);
	
    TreeNode* node = tree_node (ctree_node);
    g_return_val_if_fail (node, false);
    
    GtkCTreeNode* tnode;
    return  (node->is_moveable ()) &&
            (tnode = GTK_CTREE_NODE_NEXT (ctree_node)) &&
	    (tree_node (tnode)->is_moveable () ||
	    ((tnode = GTK_CTREE_ROW (ctree_node)->parent) &&
	     tree_node (tnode)->is_moveable ()));
}


/**
 * Hierachy::move_upwards
 *
 * Moves the currently selected tree node upwards.
 */
void Hierachy::move_upwards (GtkCTreeNode* ctree_node) {
    if (!ctree_node) ctree_node = current ();
    g_return_if_fail (ctree_node);
    
    GtkCTreeRow*  ctree_row = GTK_CTREE_ROW (ctree_node);
    g_return_if_fail (ctree_row != NULL);
    
    GtkCTreeNode* prev = GTK_CTREE_NODE_PREV (ctree_node);
    GtkCTreeRow*  prev_row = GTK_CTREE_ROW (prev);
    g_return_if_fail (prev != NULL);

    GtkCTreeNode* new_parent = prev_row->parent;
    
    gtk_clist_freeze (GTK_CLIST (widget));
    
    if (!(prev_row->is_leaf || prev_row->children)) {
    /* Empty submenu */
        gtk_ctree_move (widget, ctree_node, prev, 0);
        gtk_ctree_expand (widget, prev);
    } else if (new_parent != ctree_row->parent)
    /* Change level */
        if (prev_row->level < ctree_row->level) {
	/* Top of submenu */
            gtk_ctree_move (widget, ctree_node, new_parent, 0);
            gtk_ctree_move (widget, ctree_node, new_parent, prev);
        } else
	/* End of submenu */
            gtk_ctree_move (widget, ctree_node, new_parent, 0);
    else
    /* Default */
        gtk_ctree_move (widget, ctree_node, new_parent, prev);
	
    gtk_clist_thaw (GTK_CLIST (widget));
}

/**
 * Hierachy::move_current_downwards:
 *
 * Moves the currently selected tree node downwards.
 */
 
void Hierachy::move_downwards (GtkCTreeNode*	ctree_node) {
    if (!ctree_node) ctree_node = current ();
    g_return_if_fail (ctree_node);
    
    GtkCTreeRow*  ctree_row = GTK_CTREE_ROW (ctree_node);
    g_return_if_fail (ctree_row != NULL);
    
    gtk_clist_freeze (GTK_CLIST (widget));
    
    if (ctree_row->sibling) {
        GtkCTreeRow* sibling = GTK_CTREE_ROW (ctree_row->sibling);
	g_return_if_fail (ctree_row->sibling);
	
        if (sibling->is_leaf || 
	    (sibling->children && !sibling->expanded))
            gtk_ctree_move (widget, ctree_node,
			    sibling->parent, sibling->sibling);
        else {
	/* Begin of expanded submenu */
	    GtkCTreeNode *new_parent = ctree_row->sibling;
	    GtkCTreeNode *new_sibling = sibling->children;
	
            gtk_ctree_move (widget, ctree_node, new_parent, 0);
            gtk_ctree_move (widget, ctree_node, new_parent, new_sibling);
            gtk_ctree_expand (widget, new_parent);
	}
    } else {
    /* End of submenu */
        GtkCTreeRow *parent = GTK_CTREE_ROW (ctree_row->parent);
	g_return_if_fail (parent);
	
	GtkCTreeNode *new_parent = parent->parent;
	GtkCTreeNode *new_sibling = parent->sibling;
	
        gtk_ctree_move (widget, ctree_node, new_parent, 0);
        gtk_ctree_move (widget, ctree_node, new_parent, new_sibling);
    }
	
    gtk_clist_thaw (GTK_CLIST (widget));
}

/**
 * Hierachy::tree_select_row:
 *
 * Signal callback for tree node row selection.
 *
 * @ctree:	the widget which emited this signal
 * @row:	the selected row
 * @column:	the number of the selected column
 * @main:	our application object
 */
void Hierachy::tree_select_row (GtkCTree*	ctree,
                                GtkCTreeNode*	row,
                                gint		column,
			        Hierachy*	hierachy) {
    hierachy->current_node = row;
    Main::app.toolbar.update_widgets ();
	
    TreeNode* node;
    if ((node = hierachy->tree_node (row)) != NULL)
        Main::app.set_editor (node->get_title (),
				  node->get_description (),
	                          node->get_editor ());
}

/**
 * Hierachy::tree_unselect_row:
 *
 * Signal callback for tree node row unselection.
 *
 * @ctree:	the widget which emited this signal
 * @row:	the unselected node
 * @column:	the number of the unselected column
 * @main:	our application object
 */
void Hierachy::tree_unselect_row (GtkCTree*	ctree,
                                  GtkCTreeNode*	row,
                                  gint		column,
			          Hierachy*	hierachy) {
    TreeNode*   node;
    
    if ((node = hierachy->tree_node (row)) != NULL) {
	node->synchronize_content ();
	hierachy->update_tree_node (node, row);
    }
    hierachy->current_node = NULL;
}

/**
 * Hierachy::tree_move:
 *
 * Signal callback for tree node movement.
 *
 * @ctree:		the widget which emited this signal
 * @node:		the node moved
 * @new_parent:		the new parent node of that @node
 * @new_sibling:	the new sibling node for that @node
 * @main:		our application object
 */
void Hierachy::tree_move (GtkCTree*		ctree,
			   GtkCTreeNode*	node,
			   GtkCTreeNode*	new_parent,
			   GtkCTreeNode*	new_sibling,
			   Hierachy*		hierachy) {
    Main::app.toolbar.update_widgets ();
}


void Hierachy::load	(RootNode::Location	location,
			 const char*		custom_file) {
    gtk_clist_freeze (GTK_CLIST (widget));
    menu->load (location, custom_file);
    toolbar->load (location, custom_file);
    winopts->load (location, custom_file);
    gtk_clist_thaw (GTK_CLIST (widget));
}

void Hierachy::store	(RootNode::Location	location,
			 const char*		custom_file) {
    TreeNode *node;
    
    if ((node = tree_node (current ())) != NULL) {
	node->synchronize_content ();
	update_tree_node (node, current ());
    }
    
    menu->store (location, custom_file);
    toolbar->store (location, custom_file);
    winopts->store (location, custom_file);
}
