#ifndef __LEXER_H__
#define __LEXER_H__

#include "config.h"
#include "equal.h"

#include <iostream>
#include <hash_map>
#include <string>

IceWM_CM_BEGIN_NAMESPACE

class Lexer {
public:
    typedef hash_map<const char*, const void*, hash<const char*>, eqstr>
    					SymbolTable;

    enum Token {
        END_OF_FILE	= -1,
    	LEFT_CURLY	= '{',
    	RIGHT_CURLY	= '}',
	STRING		= 256,
	IDENTIFIER,
	SYMBOL,
	LAST
    };
    
    union Value {
        Token		token;
        const void*	symbol;
	const char*	string;
    };

    class Error : public exception {
    public:
        Error				(Lexer&		lexer,
					 Token		expection,
					 const int	state = 0,
					 const char*	description = NULL);
        virtual const char* what	(void) const { return description.c_str (); }
	
    protected:
	string				description;
    };

public:
    Lexer				(istream&	input,
    					 const char*	filename);
    virtual ~Lexer			(void);
    
    void add_symbol			(const char*	name,
					 const int	state,
    					 const void*	value);
    void add_symbol			(const char*	name,
					 const int	state,
    					 const int	id);
    int curr_line			(void) const { return line; }
    int curr_column			(void) const { return column; }
    char curr_char			(void) const { return top_char; }
    Token curr_token			(void) const { return token; }
    Value& curr_value			(void) { return value; }
    const char* get_filename		(void) { return filename; }
    
    bool eof				(void) const { return !input; }
    char get_next_char			(void);
    Token get_next_token		(void);
    bool lookup_symbol			(const char*	identifier);
    const char* get_token_name		(Token		token,
                                         const int	state);
    void set_state			(const int	new_state)
					{ state = new_state; }
    int get_state			(void) { return state; }
    
protected:
    virtual const char* whitespaces	(void) { return " \t\r\n"; }
    virtual const char* delimiters	(void) { return " \t\r\n#"; }
    virtual const char* quotation_marks	(void) { return "\""; }
    virtual const char* line_comments	(void) { return "#"; }
    virtual bool lookup_symbols		(void)
    					{ return symtab [state].size (); }
    virtual bool identifiers_as_string	(void) { return true; }
    virtual bool single_char_tokens	(void) { return true; }
    virtual bool symbol_value_as_token	(void) { return true; }
    bool keep_curr_char			(void) { return (keep_char = true); }
    
    Token				token;
    Value				value;
    hash_map<int, SymbolTable, hash<int>, eqint>
    					symtab;
    
    istream&				input;
    char				top_char;
    bool				keep_char;
    int					state;
    int					line;
    int					column;
    char*				filename;
    
    string				strbuf;
    SymbolTable::const_iterator		symbol;
};

inline Lexer::Error::Error (Lexer&		lexer,
                            Token		expection,
			    const int		state,
                            const char*		description) {
    this->description = _("** SYNTAX ERROR: ");
    
    if (lexer.get_filename ()) 
        (this->description+= lexer.get_filename ())+= ':';

    char* msg = g_strdup_printf (_("%d:%d: unexpected %s - expected: %s"),
                                 lexer.curr_line (), lexer.curr_column (),
				 lexer.get_token_name (lexer.curr_token (),
				                       state),
				 lexer.get_token_name (expection, state));
    this->description+= msg; g_free (msg);
		     
    if (description)
        ((this->description+= " (") += description)+= ')';
}
					 
inline void Lexer::add_symbol		(const char*	name,
					 const int	state,
    					 const int	id) {
    add_symbol (name, state, (const void*) id);
}

inline void Lexer::add_symbol		(const char*	name,
					 const int	state,
					 const void*	value) {
    symtab [state][name] = value;
}

inline bool Lexer::lookup_symbol	(const char*	identifier) {
    symbol = symtab [state].find (identifier);
    return (symbol != symtab [state].end ());
}

IceWM_CM_END_NAMESPACE

#endif
