#include "main.h"
#include "widgetlib/gtkhref.h"

#include <iostream>
#include <getopt.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#ifdef WITH_GDK_IMLIB
#include <gdk_imlib.h>
#endif

#include "../artwork/logo.xpm"

IceWM_CM_NAMESPACE

Main Main::app;

/**
 * Main::run:
 *
 * Starts the application. Command line parsing, initialisation of GTK+,
 * GUI construction, main loop.
 */
int Main::run (int&	argc,
               char**&	argv) {
    gtk_init (&argc, &argv);
    if (parse_args (argc, argv))
        return (1);
	
    gtk_set_locale ();
    
    gdk_rgb_init ();
#ifdef WITH_GDK_IMLIB
    gdk_imlib_init ();
#endif
    
    gtk_rc_parse (PACKAGE".gtkrc");
    build_gui ();
    gtk_main ();
    
    return 0;
}

/**
 * Main::parse_args:
 *
 * Parses command line by using getopt_long. Gives usage hints to stderr
 * on error.
 *
 * Returns	true on error.
 */
bool Main::parse_args (int&	argc,
                       char**&	argv) {
    static struct option options [] = {
    { "cfgdir",     required_argument, 0, 'c' },
    { "libdir",     required_argument, 0, 'l' },
    { 0, 0, 0, 0 }
    };
    
    cfgdir = NULL; 
    libdir = NULL; 
    
#define SET_ARGUMENT(var,name) \
	if (var) \
	    throw g_strconcat (_(name " already specified: "), var, NULL); \
	else \
	    var = g_strdup (optarg);
	    
    try {
        for (;;) {
            switch (getopt_long (argc, argv, "c:l:", options, NULL)) {
            case -1: 
                if (optind != argc)
		    throw "";
		
                if (!cfgdir)
		    cfgdir = G_DIR_SEPARATOR_S "etc" 
		             G_DIR_SEPARATOR_S "X11" 
			     G_DIR_SEPARATOR_S "icewm";
                if (!libdir)
		    libdir = PREFIX 
		             G_DIR_SEPARATOR_S "lib" 
		             G_DIR_SEPARATOR_S "X11" 
			     G_DIR_SEPARATOR_S "icewm";
                return false;
            case 'c':
	        SET_ARGUMENT (cfgdir, "Configuration directory");
		break;
            case 'l':
	        SET_ARGUMENT (libdir, "Library directory");
		break;
            default:
	        throw "";
            }
        }
    }
    catch (char *errmsg) {
        if (*errmsg) {
	    cerr << errmsg << endl;
	    g_free (errmsg);
	}
	
        cerr << "Usage: " << g_basename (*argv) 
	     << " [--cfgdir <path>] [--libdir <path>] [--nobrowser]" << endl;
        return true;
    }
    
#undef SET_ARGUMENT
}


/**
 * Main::build_gui:
 *
 * Builds the main window and it's components.
 */
void Main::build_gui (void) {
    GtkWidget* vbox;
    GtkPaned*  hpaned;

/* frame */
    frame = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_widget_set_usize (frame, 540, 320);
    gtk_window_set_title (GTK_WINDOW (frame), 
                          APPNAME " - " VERSION);
    gtk_container_set_border_width (GTK_CONTAINER (frame), 2);
    gtk_signal_connect (GTK_OBJECT (frame), "delete-event", 
                        GTK_SIGNAL_FUNC ((void (*)(...)) &quit), this);

/* vertical box: toolbar - hpaned - statusbar */
    vbox = gtk_vbox_new (FALSE, 0);
    gtk_container_add (GTK_CONTAINER (frame), vbox);
    
    gtk_box_pack_start (GTK_BOX (vbox), toolbar.init (), FALSE, TRUE, 0);
    gtk_box_pack_start (GTK_BOX (vbox), gtk_hseparator_new (), FALSE, TRUE, 0);
    
    hpaned = GTK_PANED (gtk_hpaned_new ());
    gtk_box_pack_start (GTK_BOX (vbox), GTK_WIDGET (hpaned), TRUE, TRUE, 0);    
    gtk_paned_set_position (hpaned, 166);
    
    status = GTK_STATUSBAR (gtk_statusbar_new ());
    gtk_box_pack_start (GTK_BOX (vbox), (GtkWidget*) status, FALSE, TRUE, 0);
    
/* right panel */
    editor_tank = gtk_viewport_new (0, 0);
    gtk_viewport_set_shadow_type (GTK_VIEWPORT (editor_tank), GTK_SHADOW_NONE);
    gtk_widget_ref (about_page = create_about_page ());
    
    GtkWidget* right_panel = gtk_vbox_new (FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (right_panel), 2);
    
    titlebar = gtk_viewport_new (0, 0);
    gtk_widget_set_name (titlebar, "Titlebar");
    gtk_viewport_set_shadow_type (GTK_VIEWPORT (titlebar), GTK_SHADOW_NONE);
    gtk_container_add (GTK_CONTAINER (titlebar), gtk_label_new (""));
    gtk_misc_set_alignment (GTK_MISC (GTK_BIN (titlebar)->child), 0.01, 0.5);    
    gtk_box_pack_start (GTK_BOX (right_panel), titlebar, FALSE, TRUE, 0);
    
    GtkWidget* scroller = gtk_scrolled_window_new (NULL, NULL);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scroller),
                                    GTK_POLICY_AUTOMATIC,
                                    GTK_POLICY_AUTOMATIC);
    gtk_container_add (GTK_CONTAINER (scroller), hierachy.init ());
    gtk_paned_pack1 (hpaned, scroller, TRUE, TRUE);
				    
    scroller = gtk_scrolled_window_new (NULL, NULL);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scroller),
                                    GTK_POLICY_AUTOMATIC,
                                    GTK_POLICY_AUTOMATIC);
    gtk_container_add (GTK_CONTAINER (scroller), editor_tank);
    gtk_box_pack_start (GTK_BOX (right_panel), scroller, TRUE, TRUE, 0);
    gtk_paned_pack2 (hpaned, right_panel, TRUE, TRUE);

    gtk_widget_show_all (frame);
    gtk_widget_grab_focus (hpaned->child1);
}

/**
 * Main::command_quit:
 *
 * Terminates application. TODO: Stores data if dirty and requested.
 */
void Main::quit (void) {
    gtk_main_quit ();
}

/**
 * Main::push_message:
 *
 * Puts another message to the top of the status message stack.
 *
 * @context:	the context to use
 * @message:	the message
 */
void Main::push_message (StatusContext	context,
                         const char*	message) {
    g_return_if_fail (message != NULL);
    gtk_statusbar_push (status, context, message);
}

/**s
 * Main::pop_message:
 *
 * Removes the top most message from the status message stack.
 *
 * @context:	the context to use
 */
void Main::pop_message (StatusContext context) {
    gtk_statusbar_pop (status, context);
}
    
/**
 * Main::create_about_page:
 *
 * Creates the about page.
 */
GtkWidget* Main::create_about_page (void) {
    GtkWidget* table = gtk_table_new (0, 0, FALSE);
    
    GdkPixmap* colors;
    GdkBitmap* transparancy;
    gtk_widget_realize (frame);
    colors = gdk_pixmap_create_from_xpm_d (frame->window, &transparancy, 
                                           NULL, IceWM_CM_logo_xpm);
    gtk_table_attach (GTK_TABLE (table), gtk_pixmap_new (colors, transparancy),
                      1, 2, 1, 2, GTK_FILL, GTK_FILL, 0, 0);
    gdk_bitmap_unref (transparancy);
    gdk_pixmap_unref (colors);
    
    GtkWidget* widget = gtk_label_new (APPNAME " - " VERSION);
    gtk_table_attach (GTK_TABLE (table), widget,
                      1, 2, 2, 3, GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_set_name (widget, "AppName");

    gtk_table_attach (GTK_TABLE (table), 
                      gtk_href_new ("http://icewm-cm.sourceforge.net/", NULL),
                      1, 2, 3, 4, GTK_SHRINK, GTK_FILL, 0, 0);

    GtkWidget* copyright = gtk_hbox_new (FALSE, 0);
    widget = gtk_label_new ("(C) 2000 by ");
    gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.0);
    gtk_box_pack_start (GTK_BOX (copyright), widget, FALSE, TRUE, 0);
    gtk_box_pack_start (GTK_BOX (copyright), gtk_href_new 
        ("mailto:Mathias%20Hasselmann%20<Mathias.Hasselmann@gmx.de>"
	       "?subject=[IceWM/CM%20" VERSION "]",
	 "Mathias Hasselmann <Mathias.Hasselmann@gmx.de>"),
                        FALSE, TRUE, 0);
    gtk_table_attach (GTK_TABLE (table), copyright,
                      1, 2, 4, 5, GTK_SHRINK, GTK_FILL, 0, 0);

    widget = gtk_table_new (1, 1, FALSE);
    gtk_table_attach (GTK_TABLE (widget), table,
                      1, 2, 1, 2, GTK_EXPAND, GTK_EXPAND, 0, 0);
    table = widget;

    widget = gtk_viewport_new (NULL, NULL);
        (GTK_ADJUSTMENT (gtk_adjustment_new (.5, 0, 1, 0, 0, 0)),
	 GTK_ADJUSTMENT (gtk_adjustment_new (.5, 0, 1, 0, 0, 0)));
    gtk_widget_set_name (widget, "AboutPage");
    gtk_viewport_set_shadow_type (GTK_VIEWPORT (widget), GTK_SHADOW_NONE);
    gtk_container_add (GTK_CONTAINER (widget), table);
    
    return widget;
}
 
/**
 * Main::set_editor:
 *
 * Sets the current @editor widget and it's description.
 *
 * @title:		a title for this @editor
 * @description:	a description for this editor
 * @editor:		the editor widget or NULL for about page
 */
void Main::set_editor	(const char*	title,
    			 const char*	description,
    			 GtkWidget*	new_editor) {
    g_return_if_fail (title);
    
    pop_message (CTXT_NODE_DESCRIPTION);
    if (description)
        push_message (CTXT_NODE_DESCRIPTION, description);
    
    gtk_label_set_text (GTK_LABEL (GTK_BIN (titlebar)->child), title);
    
    new_editor = new_editor ? new_editor : about_page;

    GtkWidget* old_editor = GTK_BIN (editor_tank)->child;
    if (old_editor != new_editor) {
        if (old_editor)
            gtk_container_remove (GTK_CONTAINER (editor_tank), old_editor);
				  
        gtk_container_add (GTK_CONTAINER (editor_tank), new_editor);
    }
    gtk_widget_show_all (new_editor);
}

/**
 * Main::load_icon:
 *
 * Loads a icon from IceWM's icon repository. Repeated calls for the
 * same icon are cached.
 *
 * @name:	the name IceWM name of the icon
 * returns:	the icon
 */
const TreeNode::Icon* Main::load_icon (const char* name) {
    if (!name) return NULL;

/* search the cache */
    IconCache::const_iterator icon;

    icon = icon_cache.find (name);
    if (icon != icon_cache.end ())
        return &icon->second;

/* search your disks */
    const char *prefixes [] = { 
        "", "icons" G_DIR_SEPARATOR_S, NULL };
    const char *suffixes [] = { 
        "_48x48.xpm", "_32x32.xpm", "_16x16.xpm", NULL };
    
    char* filename = find_file (name, true, true, true, 
                                NULL, prefixes, suffixes);

    if (filename) {
	TreeNode::Icon new_icon;
	    
#ifdef WITH_GDK_IMLIB
        GdkImlibImage* image = gdk_imlib_load_image (filename);
	   
	if (image) {
	    gdk_imlib_render (image, 48, 48);
	    new_icon.normal_pixmap = gdk_imlib_move_image (image);
	    new_icon.normal_bitmap = gdk_imlib_move_mask (image);
		
	    gdk_pixmap_ref (new_icon.normal_pixmap);
	    if (new_icon.normal_bitmap)
		gdk_bitmap_ref (new_icon.normal_bitmap);

	    gdk_imlib_render (image, 
		              hierachy.row_height (), 
		              hierachy.row_height ());
	    new_icon.small_pixmap = gdk_imlib_move_image (image);
	    new_icon.small_bitmap = gdk_imlib_move_mask (image);
		
	    gdk_imlib_destroy_image (image);
		
            icon_cache [name] = new_icon;
            return &icon_cache [name];
	}
#else
	new_icon.normal_pixmap = gdk_pixmap_create_from_xpm
	    (frame->window, &new_icon.normal_bitmap, NULL, filename);
		
        if (new_icon.normal_pixmap) {
	    gdk_pixmap_ref (new_icon.normal_pixmap);
	    gdk_bitmap_ref (new_icon.normal_bitmap);
	    new_icon.small_pixmap =
            new_icon.small_bitmap = NULL;
		
            icon_cache [name] = new_icon;
            return &icon_cache [name];
        }		
#endif
    }

    return NULL;
}

char* Main::find_file (const char*	filename,
		       bool		lookup_private,
		       bool		lookup_config,
		       bool		lookup_library,
		       const char*	paths [],
		       const char*	prefixes [],
		       const char*	suffixes []) {
    char* fullname;
    static struct stat buf;
    
    
    g_return_val_if_fail (filename != NULL, NULL);
    
#define TEST_FILE(name...) { \
    fullname = g_strconcat (name, NULL); \
/*g_message ("testing: %s", fullname);*/ \
    if (!stat (fullname, &buf)) return fullname; \
    g_free (fullname); \
}
    
#define TEST_PATH(path...) \
    if (prefixes) \
        if (suffixes)		/* "${path}/${prefix}${filename}${suffix}" */ \
            for (const char** prefix = prefixes; *prefix; prefix++) \
                for (const char** suffix = suffixes; *suffix; suffix++) \
                    TEST_FILE (path, G_DIR_SEPARATOR_S, \
			       *prefix, filename, *suffix) \
        else			/* "${path}/${prefix}${filename}" */ \
            for (const char** prefix = prefixes; *prefix; prefix++) \
                TEST_FILE (path, G_DIR_SEPARATOR_S, *prefix, filename) \
    else \
        if (suffixes)		/* "${path}/${filename}${suffix}" */ \
            for (const char** suffix = suffixes; *suffix; suffix++) \
                TEST_FILE (path, G_DIR_SEPARATOR_S, filename, *suffix) \
        else			/* "${path}/${filename}" */ \
            TEST_FILE (path, G_DIR_SEPARATOR_S, filename)
		    
    if (paths)
        for (const char** path = paths; *path; path++) TEST_PATH (*path);

    if (lookup_private) 
        TEST_PATH (g_get_home_dir (), G_DIR_SEPARATOR_S ".icewm");

    if (lookup_config) TEST_PATH (cfgdir);
    if (lookup_library) TEST_PATH (libdir);

#undef TEST_FILE
#undef TEST_PATH

    return NULL;
}

ostream& Main::write_header		(ostream&	os,
    					 const char*	filename) {
    g_message ("TODO: rename (\"%s\", \"%s~\") -> write_header => open_file",
    	       filename, filename);
					 
    if (!os) {
	cout << "TODO: MessageBox/g_warning-handler" << endl;
        g_warning ("Unable to write %s: %s.", filename, g_strerror (errno));
	return os;
    }

    time_t now = time (NULL);
    return os << "# " << filename << "\n"
		 "#\n"
		 "# Generated " << ctime (&now)
	      << "# by " APPNAME " - " VERSION "\n"
		 "#" << endl;
}
					 
/**
 * main:
 *
 * Guess what ;-)
 */
int main (int argc, char** argv) {
    return Main::app.run (argc, argv);
}
