#include "menuroot.h"
#include "main.h"

#include "submenu.h"
#include "program.h"
#include "separator.h"

#include <fstream>

IceWM_CM_NAMESPACE

TreeNode::InsertionInfo MenuRoot::insertion_info [] = {
    { _("Application"), _("Inserts a new application item"), 0,
      GTK_SIGNAL_FUNC (insert_new_application), &Main::app.toolbar },
    { _("Submenu"), _("Inserts a new submenu item"), 0,
      GTK_SIGNAL_FUNC (insert_new_submenu), &Main::app.toolbar },
    { _("Separator"), _("Inserts a new separator item"), 0,
      GTK_SIGNAL_FUNC (insert_new_separator), &Main::app.toolbar },
    { 0, 0, 0, 0, 0 }
};

GtkWidget* Program::top_panel = 0;
GtkWidget* Submenu::top_panel = 0;

void MenuRoot::parse_menu_content (Lexer&		lexer,
    			           GtkCTreeNode*	root_node,
                                   bool			submenu) {
    while (lexer.get_next_token () != Lexer::END_OF_FILE) {
        switch (lexer.curr_token ()) {
	case MENU:
	    parse_menu (lexer, root_node);
	    break;
	case PROG:
	    parse_program (lexer, root_node);
	    break;
	case RESTART:
	    parse_program (lexer, root_node, true);
	    break;
	case SEPARATOR:
            hierachy->insert_tree_node (new Separator (), root_node);
	    break;
	case '}':
	     if (submenu) return;
        default:
	    cerr << Lexer::Error (lexer, Lexer::SYMBOL, 0,
	                          "menu|prog|restart|separator").what () 
		 << endl;
	    break;
        }
    }
}

void MenuRoot::parse_menu (Lexer&		lexer,
    			   GtkCTreeNode*	root_node) {
    char* label = NULL;
    char* icon = NULL;
    
    try {
        if (Lexer::STRING == lexer.get_next_token ())
	    label = g_strdup (lexer.curr_value ().string);
        else
	    throw (Lexer::Error (lexer, Lexer::STRING, 0, _("menu label")));
	
        if ('-' == lexer.get_next_token ())
	    icon = NULL;
        else if (Lexer::STRING == lexer.curr_token ())
	    icon = g_strdup (lexer.curr_value ().string);
        else
	    throw (Lexer::Error (lexer, Lexer::STRING, 0, _("icon name")));

        GtkCTreeNode* new_node =
            hierachy->insert_tree_node (new Submenu (label, icon), root_node);
	
        if ('{' == lexer.get_next_token ())
	    parse_menu_content (lexer, new_node, true);
        else
	    throw (Lexer::Error (lexer, Lexer::LEFT_CURLY));
	    
    } catch (exception &e) {
        cerr << e.what () << endl;
    }	

    g_free (label);
    g_free (icon);
}

void MenuRoot::parse_program (Lexer&		lexer,
    			      GtkCTreeNode*	root_node,
			      bool		restart) {
    char*	label = NULL;
    char*	icon = NULL;
    std::string	command = "";
    
    try {
	
        if (Lexer::STRING == lexer.get_next_token ())
	    label = g_strdup (lexer.curr_value ().string);
        else
	    throw (Lexer::Error (lexer, Lexer::STRING, 0, _("program label")));
	
        if ('-' == lexer.get_next_token ())
	    icon = NULL;
        else if (Lexer::STRING == lexer.curr_token ())
	    icon = g_strdup (lexer.curr_value ().string);
        else
	    throw (Lexer::Error (lexer, Lexer::STRING, 0, _("icon name")));
/*
        while ('\n' != lexer.get_next_char () &&
	       '\r' != lexer.curr_char ())
	       command+= lexer.curr_char ();
*/
        lexer.set_state (1);
	lexer.get_next_token ();
	command = lexer.curr_value ().string;
        lexer.set_state (0);

        hierachy->insert_tree_node (new Program (label, icon, command.c_str (),
	                                         restart), root_node);
    } catch (exception &e) {
        cerr << e.what () << endl;
    }

    g_free (label);
    g_free (icon);
}

void MenuRoot::load (Location		location,
                     const char*	custom_file = 0) {
    char* filename;
    
    g_assert (hierachy != NULL);
    g_assert (ctree_node != NULL);

    if (location == LOC_CUSTOM)
        filename = g_strconcat (custom_file, G_DIR_SEPARATOR_S, 
	                        basename (), NULL);
    else 
        filename = Main::app.find_file (basename (), location <= LOC_PRIVATE,
		                                     location <= LOC_SYSTEM,
		                                     location <= LOC_DEFAULT);
    g_return_if_fail (filename != NULL);
    
    hierachy->remove_tree_node_children (this);
    
    class MenuLexer : public Lexer {
    public:
        MenuLexer			(istream&	input,
    					 const char*	filename)
          : IceWM_CM::Lexer (input, filename) { }
      
    protected:
        virtual const char* delimiters (void) {
            return (state != 1 ? " \t\r\n#{}" : "\r\n");
	}
        virtual bool single_char_tokens	(void) {
	    return (state != 1);
	}
    };
    
    ifstream in (filename);
    MenuLexer lexer (in, filename);
    
    lexer.add_symbol ("menu",		0, MENU),
    lexer.add_symbol ("prog",		0, PROG),
    lexer.add_symbol ("restart",	0, RESTART),
    lexer.add_symbol ("separator",	0, SEPARATOR);
    parse_menu_content (lexer, ctree_node);

    g_free (filename);
}

void MenuRoot::store (Location		location,
                      const char*	custom_file = 0) {
    char* filename = NULL;
    
    switch (location) {
    case LOC_PRIVATE:
        filename = g_strconcat (g_get_home_dir (),
	                        G_DIR_SEPARATOR_S ".icewm" G_DIR_SEPARATOR_S,
	                        basename (), NULL);
	break;
    case LOC_SYSTEM:
        filename = g_strconcat (Main::app.cfgdir, G_DIR_SEPARATOR_S,
	                        basename (), NULL);
	break;
    case LOC_DEFAULT:
        filename = g_strconcat (Main::app.libdir, G_DIR_SEPARATOR_S,
	                        basename (), NULL);
	break;
    case LOC_CUSTOM:
        filename = g_strconcat (custom_file, G_DIR_SEPARATOR_S, 
	                        basename (), NULL);
	break;
    }
    
    g_return_if_fail (filename && *filename);
    ofstream os (filename);
    
    Main::write_header (os, filename);
    
    for (GtkCTreeNode* node = GTK_CTREE_ROW (ctree_node)->children; node; ) {
        GtkCTreeRow* row = GTK_CTREE_ROW (node);
	
	((MenuNode*) hierachy->tree_node (node))
	                     ->store (write_ident (os, row->level - 2));
	
	if (row->children) {
	    os << ' ' << '{';
	    node = row->children;
	} else if (!row->sibling && row->parent != ctree_node) {
	    write_ident (os << endl, row->level - 3) << '}';
	    node = GTK_CTREE_ROW (row->parent)->sibling;
	} else
	    node = row->sibling;
	
	os << endl;
    }
}

void MenuRoot::insert_new_application	(GtkWidget*	widget,
    					 gpointer	user_data) {
    Main::app.hierachy.insert_interactively
        (new Program (_("<new application>"), NULL, "", false));
    Main::app.toolbar.close_menu (widget, insertion_info + 0, 
                                  _("New application"));
}

void MenuRoot::insert_new_submenu	(GtkWidget*	widget,
    					 gpointer	user_data) {
    Main::app.hierachy.insert_interactively
        (new Submenu (_("<new submenu>"), NULL));
    Main::app.toolbar.close_menu (widget, insertion_info + 1, 
                                  _("New submenu"));
}

void MenuRoot::insert_new_separator	(GtkWidget*	widget,
    					 gpointer	user_data) {
    Main::app.hierachy.insert_interactively (new Separator ());
    Main::app.toolbar.close_menu (widget, insertion_info + 2, 
                                  _("New separator"));
}
