#include "toolbar.h"
#include "main.h"
#include "hierachy.h"

#include <gtkpeeredbuttons.h>

IceWM_CM_NAMESPACE

TreeNode::InsertionInfo Toolbar::storage_menu [] = {
    { _("Private configuration"), 0, 0,
      GTK_SIGNAL_FUNC (commmand_private_config), 0 },
    { _("System defaults"), 0, 0,
      GTK_SIGNAL_FUNC (commmand_system_config), 0 },
    { _("Library defaults"), 0, 0,
      GTK_SIGNAL_FUNC (commmand_default_config), 0 },
    { _("Custom"), _("A IceWM configuration at any location"), 0,
      GTK_SIGNAL_FUNC (commmand_custom_config), 0 },
    { 0, 0, 0, 0, 0 }
};

/**
 * Toolbar::init:
 *
 * Constructs the toolbar and applies buttons.
 */
GtkWidget* Toolbar::init (void) {
    widget = GTK_TOOLBAR (gtk_toolbar_new (GTK_ORIENTATION_HORIZONTAL, 
                                           GTK_TOOLBAR_BOTH));
    gtk_widget_set_name (GTK_WIDGET (widget), "toolbar");
    gtk_toolbar_set_button_relief (widget, GTK_RELIEF_NONE);
    gtk_toolbar_set_space_style (widget, GTK_TOOLBAR_SPACE_LINE);

    insertion_info = NULL;
    insertion_menu = create_menu ();
    gtk_toolbar_append_widget (widget, insertion_menu, NULL, NULL);

    delete_button = gtk_toolbar_append_item 
        (widget, _("Delete"), _("Deletes this menu item"), NULL, NULL,
        GTK_SIGNAL_FUNC (command_delete), this);
    
    gtk_toolbar_append_space (widget); /***************************************/

    move_up_button = gtk_toolbar_append_item
        (widget, _("Up"), _("Moves this menu item upwards"), NULL, NULL,
        GTK_SIGNAL_FUNC (&command_move_up), this);
    move_down_button = gtk_toolbar_append_item 
        (widget, _("Down"), _("Moves this menu item downwards"), NULL, NULL,
        GTK_SIGNAL_FUNC (&command_move_down), this);
			     
    gtk_toolbar_append_space (widget); /***************************************/

    if (!storage_menu [0].tooltip) storage_menu [0].tooltip =
        g_strconcat (_("Your personal customization"), 
                       " (", g_get_home_dir (), "/.icewm)", NULL);
    if (!storage_menu [1].tooltip) storage_menu [1].tooltip = 
        g_strconcat (_("System-wide customized defaults"),
	               " (", Main::app.cfgdir, ")", NULL);
    if (!storage_menu [2].tooltip) storage_menu [2].tooltip =
        g_strconcat (_("Global default files"), 
	               " (", Main::app.libdir, ")", NULL);

    for (int n = 0; storage_menu [n].label; ++n)
        storage_menu [n].user_data = (void*) false;
    load_menu = create_menu ();
    fill_menu (load_menu, storage_menu, storage_menu + 0, _("Load private"),
	       _("Displays a list with configuration locations"), NULL);
    gtk_toolbar_append_widget (widget, load_menu, NULL, NULL);
    for (int n = 0; storage_menu [n].label; ++n) 
        storage_menu [n].user_data = (void*) true;
    store_menu = create_menu ();
    fill_menu (store_menu, storage_menu, storage_menu + 0, _("Store private"),
               _("Displays a list with configuration locations"), NULL);
    gtk_toolbar_append_widget (widget, store_menu, NULL, NULL);
			     
    gtk_toolbar_append_space (widget); /***************************************/

    gtk_toolbar_append_item
        (widget, _("Quit"), _("Terminates the " APPNAME), NULL, NULL,
	GTK_SIGNAL_FUNC (&command_quit), this);
	
    return GTK_WIDGET (widget);
}

/**
 * Toolbar::update_widgets:
 *
 * Updates sensitivity of movement buttons depending on the properties and
 * position of the currently selected tree node.
 *
 * @ctree:	the widget which emited this signal
 * @node:	the expanded row
 * @main:	our application object
 */
void Toolbar::update_widgets (void) {
    GtkCTreeNode*	curr;
    TreeNode*		node;

    if ((curr = Main::app.hierachy.current ()) &&
        (node = Main::app.hierachy.tree_node (curr))) {
	
	TreeNode::InsertionInfo* new_insertion_info = node->insertion_list ();
	if (insertion_info != new_insertion_info) {
	    gtk_peered_buttons_remove_all_widgets
	        (GTK_PEERED_BUTTONS (insertion_menu));
		
	    if ((insertion_info = new_insertion_info)) {
	        gchar* default_label = 
		    g_strconcat ("New ", insertion_info->label, NULL);
                fill_menu (insertion_menu, insertion_info, 
		           insertion_info + 0, default_label,
		           _("Displays a list with insertable items"), NULL);
		g_free (default_label);
	    }
	}
	
        gtk_widget_set_sensitive (delete_button, node->is_removeable ());
        gtk_widget_set_sensitive (move_up_button,
	    Main::app.hierachy.prev_node_is_moveable ());
        gtk_widget_set_sensitive (move_down_button,
	    Main::app.hierachy.next_node_is_moveable ());
        gtk_widget_set_sensitive (GTK_WIDGET (insertion_menu),
	                          node->insertion_list () != NULL);
    }
}

GtkWidget* Toolbar::create_menu (void) {
    GtkPeeredButtons *menu;

    menu = GTK_PEERED_BUTTONS
        (gtk_peered_buttons_new_with_arrow_right (NULL, GTK_SHADOW_ETCHED_IN));
    gtk_object_set_data (GTK_OBJECT (menu->box), "GtkPeeredButtons", menu);
    gtk_peered_buttons_set_relief (menu, GTK_RELIEF_NONE);
    gtk_peered_buttons_set_tooltips (menu, widget->tooltips);
    return GTK_WIDGET (menu);
}

void Toolbar::fill_menu (GtkWidget*			menu,
                         TreeNode::InsertionInfo*	items,
                         TreeNode::InsertionInfo*	default_item,
                         const char*			default_label,
                         const char*			tooltip,
                         const char*			tooltip_private) {
    if (default_item)
        gtk_peered_buttons_set_default_action
	    (GTK_PEERED_BUTTONS (menu), default_label,
	     default_item->tooltip, default_item->tooltip_private,
	     default_item->callback, default_item->user_data);

    if (tooltip)
        gtk_tooltips_set_tip (GTK_PEERED_BUTTONS (menu)->tooltips, 
	                      GTK_PEERED_BUTTONS (menu)->right,
	                      tooltip, tooltip_private);

    for (TreeNode::InsertionInfo* item = items; item && item->label; ++item)
        gtk_peered_buttons_append_item (GTK_PEERED_BUTTONS (menu), 
	                                item->label, item->tooltip,
	                                item->tooltip_private,
					item->callback, item->user_data);
}

void Toolbar::close_menu (GtkWidget*			button,
			  TreeNode::InsertionInfo*	default_item,
    			  const char*			default_label) {
    GtkPeeredButtons* menu = (GtkPeeredButtons*)
        gtk_object_get_data (GTK_OBJECT (button->parent), "GtkPeeredButtons");
	
    if (menu != NULL) {
        if (default_item)
            gtk_peered_buttons_set_default_action
                (menu, default_label ? default_label : default_item->label,
	         default_item->tooltip, default_item->tooltip_private,
		 default_item->callback, default_item->user_data);
        gtk_signal_emit_by_name (GTK_OBJECT (menu->right), "clicked");
    }
}
					 
void Toolbar::command_delete	(GtkWidget*	widget,
    				 gpointer	user_data) {
    Main::app.hierachy.remove_tree_node ();
}

void Toolbar::command_move_up	(GtkWidget*	widget,
    				 gpointer	user_data) {
    Main::app.hierachy.move_upwards ();
}

void Toolbar::command_move_down	(GtkWidget*	widget,
    				 gpointer	user_data) {
    Main::app.hierachy.move_downwards ();
}
    
void Toolbar::command_quit	(GtkWidget*	widget,
    				 gpointer	user_data) {
    Main::app.quit ();
}

void Toolbar::commmand_private_config	(GtkWidget*	widget,
    					 bool		store) {
    if (store)
        Main::app.hierachy.store (RootNode::LOC_PRIVATE);
    else
        Main::app.hierachy.load (RootNode::LOC_PRIVATE);
    
    Main::app.toolbar.close_menu (widget, storage_menu + 0,
                                  store ? _("Store private") 
				        : _("Load private"));
}

void Toolbar::commmand_system_config	(GtkWidget*	widget,
    					 bool		store) {
    if (store)
        Main::app.hierachy.store (RootNode::LOC_SYSTEM);
    else
        Main::app.hierachy.load (RootNode::LOC_SYSTEM);
    
    Main::app.toolbar.close_menu (widget, storage_menu + 1,
                                  store ? _("Store system") 
				        : _("Load system"));
}

void Toolbar::commmand_default_config	(GtkWidget*	widget,
    					 bool		store) {
    if (store)
        Main::app.hierachy.store (RootNode::LOC_DEFAULT);
    else
        Main::app.hierachy.load (RootNode::LOC_DEFAULT);
    
    Main::app.toolbar.close_menu (widget, storage_menu + 2,
                                  store ? _("Store defaults") 
				        : _("Load defaults"));
}

void Toolbar::commmand_custom_config	(GtkWidget*	widget,
    					 bool		store) {
    Main::app.toolbar.close_menu (widget, storage_menu + 3,
                                  store ? _("Store custom") 
				        : _("Load custom"));
    g_message ("TODO: Directory selector");
}
