#include "winoptroot.h"
#include "winoptnode.h"
#include "main.h"

#include <fstream>
#include <errno.h>

IceWM_CM_NAMESPACE

void WinOptRoot::load (Location		location,
                       const char*	custom_file = 0) {
    g_assert (hierachy != NULL);
    g_assert (ctree_node != NULL);
    
    if (location != LOC_CUSTOM)
        hierachy->remove_tree_node_children (this);
	
    int curr_loc = LOC_DEFAULT;
	
    do {
        char* filename;
    
        if (location == LOC_CUSTOM)
            filename = g_strconcat (custom_file, G_DIR_SEPARATOR_S, 
	                            "winoptions", NULL);
        else 
            filename = Main::app.find_file ("winoptions", 
	                                    curr_loc == LOC_PRIVATE,
		                            curr_loc == LOC_SYSTEM,
		                            curr_loc == LOC_DEFAULT);
        if (filename != NULL) {
            class WinOptLexer : public Lexer {
            public:
                WinOptLexer		(istream&	input,
    					 const char*	filename)
                  : IceWM_CM::Lexer (input, filename) { }
      
            protected:
                virtual const char* delimiters (void) {
		    return (state != 1 ? " \t\r\n#.:" : "\r\n");
		}
                virtual bool identifiers_as_string (void) {
		    return (state == 1);
		}
                virtual bool single_char_tokens	(void) {
		    return (state != 1);
		}
            };

            ifstream in (filename);
            WinOptLexer lexer (in, filename);
	    
            if (location == LOC_CUSTOM)
                hierachy->remove_tree_node_children (this);
	
	    for (unsigned n = WinOptNode::ICON; 	
	    	 n < WinOptNode::PROPERTY_COUNT; ++n) {
	        g_assert (n == WinOptNode::prop_df [n].id);
	        lexer.add_symbol (WinOptNode::prop_df [n].symbol, 0, 
		                  WinOptNode::prop_df [n].id + Lexer::LAST);
	    }

	    lexer.add_symbol ("Desktop",   1,
	                      Lexer::LAST + WinOptNode::LAYER_DESKTOP);
	    lexer.add_symbol ("Below",     1,
	                      Lexer::LAST + WinOptNode::LAYER_BELOW);
	    lexer.add_symbol ("Normal",    1,
	                      Lexer::LAST + WinOptNode::LAYER_NORMAL);
	    lexer.add_symbol ("OnTop",     1,
	                      Lexer::LAST + WinOptNode::LAYER_ON_TOP);
	    lexer.add_symbol ("Dock",      1,
	                      Lexer::LAST + WinOptNode::LAYER_DOCK);
	    lexer.add_symbol ("AboveDock", 1,
	                      Lexer::LAST + WinOptNode::LAYER_ABOVE_DOCK);
	    lexer.add_symbol ("Menu",      1,
	                      Lexer::LAST + WinOptNode::LAYER_MENU);
	    
	    while (lexer.get_next_token () != Lexer::END_OF_FILE)
	        parse_option (lexer);

            g_free (filename);
        } else if (location == LOC_CUSTOM)
	    g_warning (_("Could not parse %s: %s."), custom_file,
	                                             g_strerror (errno));
    } while (--curr_loc >= location);
}

void WinOptRoot::store (Location		location,
                      const char*	custom_file = 0) {

    char* filename = NULL;
    
    switch (location) {
    case LOC_PRIVATE:
        filename = g_strconcat (g_get_home_dir (),
	                        G_DIR_SEPARATOR_S ".icewm" G_DIR_SEPARATOR_S,
	                        "winoptions", NULL);
	break;
    case LOC_SYSTEM:
        filename = g_strconcat (Main::app.cfgdir, G_DIR_SEPARATOR_S,
	                        "winoptions", NULL);
	break;
    case LOC_DEFAULT:
        filename = g_strconcat (Main::app.libdir, G_DIR_SEPARATOR_S,
	                        "winoptions", NULL);
	break;
    case LOC_CUSTOM:
        filename = g_strconcat (custom_file, G_DIR_SEPARATOR_S, 
	                        "winoptions", NULL);
	break;
    }
    
    g_return_if_fail (filename && *filename);
    
//    ofstream os (filename);
    ostream& os = cout;

    Main::write_header (os, filename);
    
    GtkCTreeNode* class_node;
    for (class_node = GTK_CTREE_ROW (ctree_node)->children; class_node != NULL;
         class_node = GTK_CTREE_ROW (class_node)->sibling) {
	((WinOptNode*) hierachy->tree_node (class_node))->store (os);
	 
        GtkCTreeNode* window_node;
        for (window_node = GTK_CTREE_ROW (class_node)->children; 
	     window_node != NULL;
             window_node = GTK_CTREE_ROW (window_node)->sibling)
	    ((WinOptNode*) hierachy->tree_node (window_node))->store (os);
    }	    
}

void WinOptRoot::parse_option		(Lexer&		lexer) {
    PropertyNode::PropValue value;
    char*	wm_class = NULL;
    char*	wm_name = NULL;
    int		option;
	
    try {
        if (lexer.curr_token () == Lexer::IDENTIFIER)
	    wm_class = g_strdup (lexer.curr_value ().string);
	else
	    throw (Lexer::Error (lexer, Lexer::IDENTIFIER, 0, _("class name")));

        if (lexer.get_next_token () != '.')
	    throw (Lexer::Error (lexer, (Lexer::Token) '.', 0, _("dot")));
	    
        if (lexer.get_next_token () == Lexer::IDENTIFIER) {
	    wm_name = wm_class;
	    wm_class = g_strdup (lexer.curr_value ().string);
	    
            if (lexer.get_next_token () != '.')
	        throw (Lexer::Error (lexer, (Lexer::Token) '.', 0, _("dot")));
		
	    lexer.get_next_token ();
	}
	
        option = (int) lexer.curr_token () - Lexer::LAST;
	if (option < WinOptNode::ICON || 
	    option >= WinOptNode::PROPERTY_COUNT)
	    throw (Lexer::Error (lexer, Lexer::IDENTIFIER, 0, 
	                         _("window option, see IceWM docs for details")));
				 
        if (lexer.get_next_token () != ':')
	    throw (Lexer::Error (lexer, (Lexer::Token) ':', 0, _("colon")));
	
	lexer.set_state (1);
	
	if (option == WinOptNode::LAYER) {
	    int layer = lexer.get_next_token () - Lexer::LAST;
	    if (layer >= WinOptNode::LAYER_DESKTOP &&
	        layer <= WinOptNode::LAYER_MENU)
		value.integer = layer;
	    else
	        value.integer = atoi (lexer.curr_value ().string);
        } else {
	    lexer.get_next_token ();
	    switch (WinOptNode::prop_df [option].type) {
	    case PropertyNode::PT_INTEGER:
	        value.integer = atoi (lexer.curr_value ().string);
	        break;
	    case PropertyNode::PT_BOOLEAN:
	        value.boolean = atoi (lexer.curr_value ().string);
	        break;
	    case PropertyNode::PT_STRING:
	        value.string = g_strdup (lexer.curr_value ().string);
	        break;
	    default:
	        g_assert (false);
	    }
	}
	
	lexer.set_state (0);
	
	GtkCTreeNode* node;
	if ((node = find_class (wm_class)) == NULL)
	    node = hierachy->insert_tree_node (new WinOptNode (wm_class),
	                                       ctree_node);
        if (wm_name) {
	    GtkCTreeNode* class_node = node;
	    if ((node = find_window (class_node, wm_name)) == NULL)
	        node = hierachy->insert_tree_node (new WinOptNode (wm_class, 
		                                                   wm_name),
					           class_node);
	}
	
	PropertyNode* prop_node = (PropertyNode*) hierachy->tree_node (node);
	switch (WinOptNode::prop_df [option].type) {
	case PropertyNode::PT_INTEGER:
	    prop_node->set_property (option, value.integer);
	    break;
	case PropertyNode::PT_BOOLEAN:
	    prop_node->set_property (option, value.boolean);
	    break;
	case PropertyNode::PT_STRING:
	    prop_node->set_property (option, value.string);
	    if (option == WinOptNode::ICON)
	        Main::app.hierachy.update_tree_node (node);
	    g_free (value.string);
	    break;
	}
    } catch (exception &e) {
        cerr << e.what () << endl;
    }

    g_free (wm_class);
    g_free (wm_name);
}

GtkCTreeNode* WinOptRoot::find_class	(const char*	wm_class) {
    GtkCTreeNode* node;

    g_assert (ctree_node);
    g_assert (hierachy);
    g_assert (wm_class);
    
    for (node = GTK_CTREE_ROW (ctree_node)->children; node != NULL && 
         strcmp (hierachy->tree_node (node)->get_label (), wm_class);
         node = GTK_CTREE_ROW (node)->sibling);
	 
    return node;
}

GtkCTreeNode* WinOptRoot::find_window	(GtkCTreeNode*	class_node,
    					 const char*	wm_name) {
    GtkCTreeNode* node = NULL;

    g_assert (ctree_node);
    g_assert (hierachy);
    g_assert (wm_name);
    
    if (class_node != NULL)
        for (node = GTK_CTREE_ROW (class_node)->children; node != NULL && 
	     strcmp (hierachy->tree_node (node)->get_label (), wm_name);
             node = GTK_CTREE_ROW (node)->sibling);
	       
    return node;
}

TreeNode::InsertionInfo* WinOptRoot::insertion_list
    					(void) {
    return WinOptNode::insertion_info;
}
